/*
 * Copyright (c) 2015 NIBIO <http://www.nibio.no/>. 
 * 
 * This file is part of RoughageNutritionModel.
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 * 
 */

package no.bioforsk.vips.model.roughagenutritionmodel;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;
import no.nibio.vips.entity.ModelConfiguration;
import no.nibio.vips.entity.Result;
import no.nibio.vips.entity.ResultImpl;
import no.nibio.vips.entity.WeatherObservation;
import no.nibio.vips.model.ConfigValidationException;
import no.nibio.vips.model.Model;
import no.nibio.vips.model.ModelExcecutionException;
import no.nibio.vips.model.ModelId;
import no.nibio.vips.util.InvalidAggregationTypeException;
import no.nibio.vips.util.WeatherElements;
import no.nibio.vips.util.WeatherObservationListException;
import no.nibio.vips.util.WeatherUtil;


/**
 * @copyright 2015 <a href="http://www.nibio.no/">NIBIO</a>
 * @author Tor-Einar Skog <tor-einar.skog@nibio.no>
 */
public class RoughageNutritionModel implements Model {
    
    public final static ModelId MODEL_ID = new ModelId("ROUGHAGENU");
    
    private TimeZone timeZone;
    
    public final static int JORDTYPE_LEIR = 1;
    public final static int JORDTYPE_SAND = 2;
    public final static int JORDTYPE_SILT = 3;

    public final static int ENGSAMMENSETNING_KLOEVERANDEL_LITEN = 1;
    public final static int ENGSAMMENSETNING_KLOEVERANDEL_MEDIUM = 2;
    public final static int ENGSAMMENSETNING_KLOEVERANDEL_STOR = 3;
    
    private List<WeatherObservation> TM;
    private List<WeatherObservation> TJM10;
    private List<WeatherObservation> RR;
    private List<WeatherObservation> EPP;
    private List<WeatherObservation> Q0;
    private List<WeatherObservation> TMD;
    private List<WeatherObservation> TJM10D;
    private List<WeatherObservation> RRD;
    private List<WeatherObservation> Q0D;
    private List<WeatherObservation> EPPD;
    
    private Date firstHarvest;
    private Date secondHarvest;
    private Date firstPossibleGrowthStartDate;
    private Integer soilType;
    private Integer cloverShare;
    private List<OptimizationObservation> optimizationObservations;
    
    
    
    @Override
    public List<Result> getResult() throws ModelExcecutionException {
        RoughageNutritionModelImpl impl = new RoughageNutritionModelImpl();
        ResultMatrix resultMatrix = impl.beregnModell(
                TMD, 
                TJM10D, 
                RRD, 
                EPPD, 
                Q0D, 
                firstHarvest, 
                secondHarvest, 
                firstPossibleGrowthStartDate,
                soilType, 
                cloverShare, 
                optimizationObservations, 
                timeZone
        );
        
        List<Result> results = new ArrayList<>();
                
        // Get all labels
        //Set labels = resultMatrix.getValueLabelsForDate(aprilFirst);
        Date resultDate = firstPossibleGrowthStartDate;
        // Iterate all dates in matrix, starting on the firstPossibleGrowthStartDate
        Calendar cal = Calendar.getInstance(this.timeZone);
        while(resultMatrix.getValueLabelsForDate(resultDate) != null)
        {
            Result result = new ResultImpl();
            result.setValidTimeStart(resultDate);
            for(String label:resultMatrix.getValueLabelsForDate(resultDate))
            {
                result.setValue(this.getModelId().toString(), label, resultMatrix.getDefaultFormattedDoubleValue(resultDate, label));
            }
            cal.setTime(resultDate);
            cal.add(Calendar.DATE, 1);
            resultDate = cal.getTime();
            results.add(result);
        }
        
        return results;
    }
    
    
    

    @Override
    public ModelId getModelId() {
        return RoughageNutritionModel.MODEL_ID;
    }

    @Override
    public String getModelName() {
        return this.getModelName(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelName(String language) {
        return "Grovfôrmodellen";
    }

    @Override
    public String getLicense() {
        return "GNU Affero General Public License v3";
    }

    @Override
    public String getCopyright() {
        return "(c) 2018 NIBIO";
    }

    @Override
    public String getModelDescription() {
        return this.getModelDescription(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelDescription(String language) {
        return "Se README";
    }

    @Override
    public String getWarningStatusInterpretation() {
        return this.getWarningStatusInterpretation(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getWarningStatusInterpretation(String language) {
        return "TODO";
    }

    @Override
    public String getModelUsage() {
        return this.getModelName(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelUsage(String language) {
        return "TODO";
    }

    @Override
    public String getSampleConfig() {
        return  "{\n" +
                "\t\"loginInfo\":{\n" +
                "\t\t\"username\":\"example\",\n" +
                "\t\t\"password\":\"example\"\n" +
                "\t},\n" +
                "\t\"modelId\":\"" + MODEL_ID.toString() + "\",\n" +
                "\t\"configParameters\":{\n" +
                "\t\t\"timeZone\":\"Europe/Oslo\",\n" +
                "\t\t\"firstHarvest\":\"2015-06-01\",\n" +
                "\t\t\"secondHarvest\":\"2015-08-01\",\n" +
                "\t\t\"soilType\":1,\n" +
                "\t\t\"cloverShare\":2,\n" +
                "\t\t\"observations\":[\n" +
                "\t\t{\n" +
                "\t\t\t\t\"timeMeasured\": \"2015-04-01T00:00:00+02:00\",\n" +
                "\t\t\t\t\"elementMeasurementTypeId\":\"TM\",\n" +
                "\t\t\t\t\"logIntervalId\":2,\n" + 
                "\t\t\t\t\"value\":10.9\n" +
                "\t\t},\n" +
                "\t\t{\n" +
                "\t\t\t\t\"timeMeasured\": \"2012-04-01T00:00:00+02:00\",\n" +
                "\t\t\t\t\"elementMeasurementTypeId\":\"TJM10\",\n" +
                "\t\t\t\t\"logIntervalId\":2,\n" +
                "\t\t\t\t\"value\":2.9\n" +
                "\t\t},\n" +
                "\t\t{\n" +
                "\t\t\t\t\"timeMeasured\": \"2012-04-01T00:00:00+02:00\",\n" +
                "\t\t\t\t\"elementMeasurementTypeId\":\"RR\",\n" +
                "\t\t\t\t\"logIntervalId\":2,\n" + 
                "\t\t\t\t\"value\":1.7\n" +
                "\t\t},\n" +
                "\t\t{\n" +
                "\t\t\t\t\"timeMeasured\": \"2012-04-25T00:00:00+02:00\",\n" +
                "\t\t\t\t\"elementMeasurementTypeId\":\"EPP\",\n" +
                "\t\t\t\t\"logIntervalId\":2,\n" + 
                "\t\t\t\t\"value\":0.5\n" +
                "\t\t},\n" +
                "\t\t{\n" +
                "\t\t\t\t\"timeMeasured\": \"2012-04-01T00:00:00+02:00\",\n" +
                "\t\t\t\t\"elementMeasurementTypeId\":\"Q0\",\n" +
                "\t\t\t\t\"logIntervalId\":2,\n" + 
                "\t\t\t\t\"value\":10.7\n" +
                "\t\t},\n" +
                "\t\t]\n" +
                "\t}\n" +
                "}\n";
    }

    @Override
    public void setConfiguration(ModelConfiguration config) throws ConfigValidationException {

        /*
            // Daily values
            List<WeatherObservation> luftTemperatur, TM
            List<WeatherObservation> jordTemperatur, TJM10
            List<WeatherObservation> nedboer, RR
            List<WeatherObservation> potensiellFordamping, EPP
            List<WeatherObservation> globalStraaling, Q0
            Date foersteslaatt, 
            Date andreslaatt, 
            Date tidligsteVekststart, // Default: 1. mars
            int jordtype, 
            int kloeverandel, 
            List<OptimizationObservation> observasjoner
        
        
        */
        // Hourly data
        this.TM = new ArrayList<>();
        this.TJM10 = new ArrayList<>();
        this.RR = new ArrayList<>();
        this.EPP = new ArrayList<>();
        this.Q0 = new ArrayList<>();
        
        this.TMD = new ArrayList<>();
        this.TJM10D = new ArrayList<>();
        this.RRD = new ArrayList<>();
        this.EPPD = new ArrayList<>();
        this.Q0D = new ArrayList<>();
        
        ObjectMapper mapper = new ObjectMapper();
        
        // Set time zone
        if(config.getConfigParameter("timeZone") != null)
        {
            this.timeZone = TimeZone.getTimeZone((String) config.getConfigParameter("timeZone"));
        }
        else
        {
            throw new ConfigValidationException("Missing timezone information");
        }
        if(config.getConfigParameter("firstHarvest") != null)
        {
            this.firstHarvest = mapper.convertValue(config.getConfigParameter("firstHarvest"), Date.class);
        }
        else
        {
            throw new ConfigValidationException("Missing date of first harvest");
        }
        if(config.getConfigParameter("secondHarvest") != null)
        {
            this.secondHarvest = mapper.convertValue(config.getConfigParameter("secondHarvest"), Date.class);
        }
          
        if(config.getConfigParameter("firstPossibleGrowthStartDate") != null)
        {
            this.firstPossibleGrowthStartDate = mapper.convertValue(config.getConfigParameter("firstPossibleGrowthStartDate"), Date.class);
        }
        else
        {
            Calendar cal = Calendar.getInstance();
            cal.setTime(firstHarvest);
            cal.set(Calendar.MONTH, Calendar.MARCH);
            cal.set(Calendar.DATE, 1);
            this.firstPossibleGrowthStartDate = cal.getTime();
        }
        
        if(config.getConfigParameter("soilType") != null)
        {
            this.soilType = mapper.convertValue(config.getConfigParameter("soilType"), Integer.class);
        }
        if(config.getConfigParameter("cloverShare") != null)
        {
            this.cloverShare = mapper.convertValue(config.getConfigParameter("cloverShare"), Integer.class);
        }
        if(config.getConfigParameter("optimizationInfo") != null)
        {
            List<String> optimizationInfo = mapper.convertValue(config.getConfigParameter("optimizationInfo"), new TypeReference<List<String>>(){});
            /*for(String inf:optimizationInfo)
            {
                System.out.println(inf);
            }*/
            this.optimizationObservations = this.parseOptimizationInfo(optimizationInfo, this.timeZone);
        }
        
        
        
        
        
        List<WeatherObservation> observations = mapper.convertValue(config.getConfigParameter("observations"), new TypeReference<List<WeatherObservation>>(){});
        
        for(WeatherObservation o:observations)
        {
            //System.out.println(o);
            switch(o.getElementMeasurementTypeId())
            {
                case WeatherElements.TEMPERATURE_MEAN:
                    if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1H))
                    {
                        TM.add(o);
                    }
                    else if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1D))
                    {
                        TMD.add(o);
                    }
                    break;
                case WeatherElements.SOIL_TEMPERATURE_10CM_MEAN:
                    if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1H))
                    {
                        TJM10.add(o);
                    }
                    else if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1D))
                    {
                        TJM10D.add(o);
                    }
                    break;
                case WeatherElements.PRECIPITATION:
                    if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1H))
                    {
                        RR.add(o);
                    }
                    else if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1D))
                    {
                        RRD.add(o);
                    }
                    break;
                case WeatherElements.POTENTIAL_EVAPORATION:
                    if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1H))
                    {
                        EPP.add(o);
                    }
                    else if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1D))
                    {
                        EPPD.add(o);
                    }
                    break;
                case WeatherElements.GLOBAL_RADIATION:
                    if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1H))
                    {
                        Q0.add(o);
                    }
                    else if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1D))
                    {
                        Q0D.add(o);
                    }
                    break;
                default:
                    // Keep calm and continue importing data
                    break;
            }
        }
        // Fallback: Using hourly values to calculate dailies
        if(TMD == null || TMD.isEmpty())
        {
            try {
                List<WeatherObservation> dailyMeanTemperatures = new WeatherUtil().getAggregatedDailyValues(TM,
                        this.timeZone,
                        15,
                        WeatherUtil.AGGREGATION_TYPE_AVERAGE, 1);
                if(dailyMeanTemperatures != null)
                {
                    this.TMD = dailyMeanTemperatures;
                    
                }
            } catch (WeatherObservationListException | InvalidAggregationTypeException ex) {
                throw new ConfigValidationException(ex.getMessage());
            }
        }
        if(TJM10D == null || TJM10D.isEmpty())
        {
            try {
                List<WeatherObservation> dailyMeanSoilTemperatures = new WeatherUtil().getAggregatedDailyValues(TJM10,
                        this.timeZone,
                        15,
                        WeatherUtil.AGGREGATION_TYPE_AVERAGE, 1);
                if(dailyMeanSoilTemperatures != null)
                {
                    this.TJM10D = dailyMeanSoilTemperatures;
                    
                }
            } catch (WeatherObservationListException | InvalidAggregationTypeException ex) {
                throw new ConfigValidationException(ex.getMessage());
            }
        }
        if(RRD == null || RRD.isEmpty())
        {
            try {
                List<WeatherObservation> dailyPrecipitation = new WeatherUtil().getAggregatedDailyValues(RR,
                        this.timeZone,
                        15,
                        WeatherUtil.AGGREGATION_TYPE_SUM, 1);
                if(dailyPrecipitation != null)
                {
                    this.RRD = dailyPrecipitation;
                    
                }
            } catch (WeatherObservationListException | InvalidAggregationTypeException ex) {
                throw new ConfigValidationException(ex.getMessage());
            }
        }
        
        if(EPPD == null || EPPD.isEmpty())
        {
            try {
                List<WeatherObservation> dailyEvap = new WeatherUtil().getAggregatedDailyValues(EPP,
                        this.timeZone,
                        15,
                        WeatherUtil.AGGREGATION_TYPE_SUM, 1);
                if(dailyEvap != null)
                {
                    this.EPPD = dailyEvap;
                    
                }
            } catch (WeatherObservationListException | InvalidAggregationTypeException ex) {
                throw new ConfigValidationException(ex.getMessage());
            }
        }
        
        if(Q0D == null || Q0D.isEmpty())
        {
            try {
                List<WeatherObservation> dailyRad = new WeatherUtil().getAggregatedDailyValues(Q0,
                        this.timeZone,
                        15,
                        WeatherUtil.AGGREGATION_TYPE_SUM_GLOBAL_RADIATION, 1);
                if(dailyRad != null)
                {
                    this.Q0D = dailyRad;
                    
                }
            } catch (WeatherObservationListException | InvalidAggregationTypeException ex) {
                throw new ConfigValidationException(ex.getMessage());
            }
        }
    }

    private List<OptimizationObservation> parseOptimizationInfo(List<String> optimizationInfo, TimeZone timeZone) throws ConfigValidationException {
        List<OptimizationObservation> retVal = new ArrayList<>();
        SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd");
        format.setTimeZone(timeZone);
        for(String line:optimizationInfo)
        {
            // The -1 as second parameter ensures that trailing delimiters will not be discarded,
            // Thus avoiding ArrayIndexOutOfBoundsException when parsing
            // E.g. "2015-06-02,2,3,,,,"
            String[] parts = line.split(",",-1); 
            try
            {
                OptimizationObservation optObs = new OptimizationObservation(
                        format.parse(parts[0]), // Date
                        (! parts[1].trim().isEmpty() ? Double.parseDouble(parts[1]) : null), // Height
                        (! parts[2].trim().isEmpty() ? Double.parseDouble(parts[2]) : null), // MSC
                        (! parts[4].trim().isEmpty() ? Double.parseDouble(parts[4]) : null), // NDF
                        (! parts[5].trim().isEmpty() ? Double.parseDouble(parts[5]) : null), // INDF
                        (! parts[6].trim().isEmpty() ? Double.parseDouble(parts[6]) : null), // Raw protein
                        (! parts[3].trim().isEmpty() ? Double.parseDouble(parts[3]) : null) // FEm
                );
                if(optObs.getHeight() == null && optObs.getMSC() == null)
                {
                    throw new ConfigValidationException("Optimization error: Height and/or MSC must be set");
                }
               retVal.add(optObs);
            }
            catch(ArrayIndexOutOfBoundsException | ParseException ex) {
                throw new ConfigValidationException("There's something wrong with this line of optimization observations: " + 
                        line + " . The error is: " + ex.getClass() + ":" + ex.getMessage());
            }
        }
        return retVal;
    }

}
